/***********************************************************************
 * Copyright (c) 2003 University of Kent.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.cs.kent.ac.uk/projects/ocl/CPL.html
 ***********************************************************************/

package uk.ac.kent.cs.ocl20.standard.lib;

public class OclRealImpl extends OclAnyImpl implements OclReal {
	protected OclRealImpl(double r, StdLibAdapter adapter) {
		super(adapter);
		_implementation = r;
	}
	double _implementation;
	public double implementation() {
		return _implementation;
	}

	public OclBoolean equalTo(OclReal r2) {
		return adapter.Boolean(this.implementation() == ((OclRealImpl) r2).implementation());
	}

	public OclBoolean notEqualTo(OclReal r2) {
		return adapter.Boolean(this.implementation() != ((OclRealImpl) r2).implementation());
	}
	
	public Object asJavaObject() {
		return new Double(implementation());
	}

	//--- IOclReal ---
	public OclReal add(OclReal r2) {
		return adapter.Real(this.implementation() + ((OclRealImpl) r2).implementation());
	}

	public OclReal subtract(OclReal r2) {
		return adapter.Real(this.implementation() - ((OclRealImpl) r2).implementation());
	}

	public OclReal multiply(OclReal r2) {
		return adapter.Real(this.implementation() * ((OclRealImpl) r2).implementation());
	}

	public OclReal negate() {
		return adapter.Real(this.implementation() * -1);
	}

	public OclReal divide(OclReal r2) {
		if (adapter.impl(r2).doubleValue() == 0) return adapter.Undefined();
		return adapter.Real(this.implementation() / ((OclRealImpl) r2).implementation());
	}

	public OclReal abs() {
		return adapter.Real(Math.abs(this.implementation()));
	}

	public OclInteger floor() {
		return adapter.Integer((int) Math.floor(this.implementation()));
	}

	public OclInteger round() {
		return adapter.Integer((int) Math.round(this.implementation()));
	}

	public OclReal max(OclReal r2) {
		return adapter.Real(Math.max(this.implementation(), ((OclRealImpl) r2).implementation()));
	}

	public OclReal min(OclReal r2) {
		return adapter.Real(Math.min(this.implementation(), ((OclRealImpl) r2).implementation()));
	}

	public OclBoolean lessThan(OclReal r2) {
		return adapter.Boolean(this.implementation() < ((OclRealImpl) r2).implementation());
	}

	public OclBoolean greaterThan(OclReal r2) {
		return adapter.Boolean(this.implementation() > ((OclRealImpl) r2).implementation());
	}

	public OclBoolean lessThanOrEqualTo(OclReal r2) {
		return adapter.Boolean(this.implementation() <= ((OclRealImpl) r2).implementation());
	}

	public OclBoolean greaterThanOrEqualTo(OclReal r2) {
		return adapter.Boolean(this.implementation() >= ((OclRealImpl) r2).implementation());
	}

	//--- IOclVisitable ---
	public Object accept(OclVisitor v, Object obj) {
		return v.visit(this, obj);
	}

	//---Object---
	public String toString() {
		return String.valueOf(implementation());
	}
	public boolean equals(Object r) {
		if (r instanceof OclRealImpl)
			return adapter.impl(this.equalTo((OclRealImpl) r)).booleanValue();
		return false;
	}

	public int hashCode() {
		return (int) implementation();
	}

	public Object clone() {
		return new OclRealImpl(implementation(), super.adapter);
	}
}
